import gitty, sys
from . import aliases, importer
from . types.defaults import FIELD_TYPES
from .. animation import runner
from .. import data_maker
from .. layout.geometry import gen_matrix
from .. layout.multimap import MultiMapBuilder
from .. util import files
import copy
from .. import log
import traceback

RESERVED_PROPERTIES = 'name', 'data'


def _make_object(*args, field_types=FIELD_TYPES, **kwds):
    return importer.make_object(*args, field_types=field_types, **kwds)


def _make_layout(layout, driver=None, drivers=None, maker=None):
    if driver is None and drivers is None:
        raise ValueError('Projects has no driver or drivers section')

    coord_map = layout.pop('coordMap', None)

    if drivers is None:
        drivers = [_make_object(**driver)]
    else:
        build = MultiMapBuilder()

        def make_driver(width, height, matrix=None, **kwds):
            build.addRow(gen_matrix(width, height, **(matrix or {})))
            return _make_object(width=width, height=height, **kwds)

        if driver:
            # driver is a default for each driver.
            drivers = [dict(driver, **d) for d in drivers]

        drivers = [make_driver(**d) for d in drivers]
        coord_map = coord_map or build.map

    maker = data_maker.Maker(**(maker or {}))
    return _make_object(drivers, coordMap=coord_map, maker=maker, **layout)


def make_animation(layout, animation, run=None):
    reserved = {p: animation.pop(p, None) for p in RESERVED_PROPERTIES}
    animation = _make_object(layout, **animation)

    # Add the reserved properties back in.
    for k, v in reserved.items():
        (v is not None) and setattr(animation, k, v)

    animation.set_runner(runner.Runner(**(run or {})))
    return animation


def _make_project(path=None, animation=None, run=None, **kwds):
    gitty.sys_path.extend(path)
    layout = _make_layout(**kwds)

    return make_animation(layout, animation or {}, run or {})


def project_to_animation(desc, default):
    project = aliases.resolve(default or {}, desc)
    return _make_project(**project)



import argparse, gitty, os, sys
from . import common_flags
from .. import log
from .. project.importer import import_symbol
from .. project.preset_library import PresetLibrary

__all__ = ['main']
COMMANDS = (
    'all_pixel', 'clear_cache', 'color', 'devices', 'demo', 'run', 'update')
MODULES = {c: import_symbol('.' + c, 'bibliopixel.main') for c in COMMANDS}


def no_command(*_):
    print('ERROR: No command entered')
    print('Valid:', ', '.join(COMMANDS))
    return -1


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    for name, module in sorted(MODULES.items()):
        subparser = subparsers.add_parser(name, help=module.__doc__)
        module.set_parser(subparser)

    common_flags.add_common_flags(parser)

    argv = ['--help' if i == 'help' else i for i in sys.argv[1:]]

    try:
        argv.remove('--version')
    except:
        pass
    else:
        print('BiblioPixel version %s' % common_flags.VERSION)
        if not argv:
            return
    args = parser.parse_args(argv)

    try:
        log.set_log_level(args.loglevel)
        presets = common_flags.ENABLE_PRESETS and PresetLibrary(
            os.path.expanduser(args.presets), True)

        run = getattr(args, 'run', no_command)
        result = run(args, presets) or 0
    except Exception as e:
        if args.verbose:
            raise
        print('ERROR:', e.args[0], file=sys.stderr)
        print(*e.args[1:], sep='\n', file=sys.stderr)
        result = getattr(e, 'errorcode', -1)

    sys.exit(result)



import math, threading, time

from .. import colors, data_maker, font, matrix, log
from . layout import Layout
from . geometry.matrix import Rotation, gen_matrix, pixel_positions_from_matrix


class Matrix(Layout):

    def __init__(self, drivers, width=0, height=0, coordMap=None,
                 rotation=Rotation.ROTATE_0, vert_flip=False,
                 serpentine=True,
                 threadedUpdate=False, brightness=255,
                 pixelSize=(1, 1), **kwargs):
        """Main class for matricies.
        driver - instance that inherits from DriverBase
        width - X axis size of matrix
        height - Y axis size of matrix
        coordMap - a 2D matrix defining the X,Y to strip index mapping. Not needed in most cases
        rotation - how to rotate when generating the map. Not used if coordMap specified
        vert_flip - flips the generated map along the Y axis. This along with rotation can achieve any orientation
        """
        super().__init__(drivers, threadedUpdate, brightness,
                         maker=kwargs.get('maker', data_maker.MAKER))

        self.width = width or getattr(self.drivers[0], 'width') or 32
        self.height = height or getattr(self.drivers[0], 'height') or 32
        self.pixelSize = pixelSize
        pw, ph = self.pixelSize

        # if both are 0 try to assume it's a square display
        if self.width == 0 and self.height == 0:
            square = int(math.sqrt(self.numLEDs))
            if (square * square) == self.numLEDs:
                self.width = self.height = square
            else:
                raise TypeError(
                    "No width or height passed but the number of LEDs is not a perfect square")

        if self.width * self.height > self.numLEDs:
            raise ValueError('width * height cannot exceed total pixel count!'
                             ' %s * %s > %s'
                             % (self.width, self.height, self.numLEDs))

        if coordMap:
            self.matrix_map = coordMap
        else:
            if len(self.drivers) == 1:
                log.info('Auto generating coordinate map. Use gen_matrix directly if more control needed.')
                self.matrix_map = gen_matrix(self.width, self.height,
                                             serpentine=serpentine,
                                             rotation=rotation,
                                             y_flip=vert_flip)
            else:
                raise TypeError(
                    "Must provide coordMap if using multiple drivers!")

        self.set_pixel_positions(pixel_positions_from_matrix(self.matrix_map))

        # if 90 or 270 rotation dimensions need to be swapped so they match the
        # matrix rotation
        if rotation % 2 != 0:
            w = self.width
            h = self.height
            self.width = h
            self.height = w

        self.texture = None
        self.set = self._setColor

        if pw < 0 or pw > self.width or ph < 0 or ph > self.height:
            raise ValueError(
                "pixelSize must be greater than 0 and not larger than total matrix!")
        if self.width % pw != 0 or self.height % ph != 0:
            raise ValueError(
                "pixelSize must evenly divide into matrix dimensions!")

        if pw == 1 and ph == 1:
            self._set = self.__setNormal
        else:
            self._set = self.__setScaled
            self.width = self.width / pw
            self.height = self.height / ph
            self.numLEDs = self.width * self.height

        self.fonts = font.fonts

    def get_pixel_positions(self):
        return pixel_positions_from_matrix(self.coord_map)

    def loadFont(self, name, height, width, data):
        self.fonts[name] = {
            "data": data,
            "height": height,
            "width": width
        }

    def setTexture(self, tex=None):
        if tex is None:
            self.texture = tex
            self.set = self._setColor
        else:
            if not isinstance(tex, list):
                raise ValueError("Texture must be a list!")
            elif len(tex) != self.height:
                raise ValueError(
                    "Given texture is must be {} high!".format(self.height))
            else:
                for r in tex:
                    if not isinstance(r, list):
                        raise ValueError("Texture rows must be lists!")
                    elif len(r) != self.width:
                        raise ValueError(
                            "Texture rows must be {} wide!".format(self.width))

            self.texture = tex
            self.set = self._setTexture

    def __setNormal(self, x, y, color):
        try:
            pixel = self.matrix_map[y][x]
            self._set_base(pixel, color)
        except IndexError:
            pass

    def __setScaled(self, x, y, color):
        sx = x * self.pixelSize[0]
        sy = y * self.pixelSize[1]

        for xs in range(sx, sx + self.pixelSize[0]):
            for ys in range(sy, sy + self.pixelSize[1]):
                self.__setNormal(xs, ys, color)

    # Set single pixel to Color value
    def _setColor(self, x, y, color=None):
        try:
            self._set(x, y, color or (0, 0, 0))
        except IndexError:
            pass

    def _setTexture(self, x, y, color=None):
        if x >= 0 and y >= 0:
            try:
                self._set(x, y, color or self.texture[y][x])
            except IndexError:
                pass

    def get(self, x, y):
        try:
            pixel = self.matrix_map[y][x]
            return self._get_base(pixel)
        except IndexError:
            return 0, 0, 0

    def setHSV(self, x, y, hsv):
        color = colors.hsv2rgb(hsv)
        self._set(x, y, color)

    def setRGB(self, x, y, r, g, b):
        color = (r, g, b)
        self._set(x, y, color)

    ##########################################################################
    # Drawing Functions
    # Lovingly borrowed from Adafruit
    # https://github.com/adafruit/Adafruit-GFX-Library/blob/master/Adafruit_GFX.cpp
    ##########################################################################

    def drawCircle(self, x0, y0, r, color=None):
        """Draws a circle at point x0, y0 with radius r of the specified RGB color"""
        matrix.draw_circle(self.set, x0, y0, r, color)

    def fillCircle(self, x0, y0, r, color=None):
        """Draws a filled circle at point x0,y0 with radius r and specified color"""
        matrix.fill_circle(self.set, x0, y0, r, color)

    def drawLine(self, x0, y0, x1, y1, color=None, colorFunc=None, aa=False):
        matrix.draw_line(self.set, x0, y0, x1, y1, color, colorFunc, aa)

    # Bresenham's algorithm
    def bresenham_line(self, x0, y0, x1, y1, color=None, colorFunc=None):
        """Draw line from point x0,y0 to x,1,y1. Will draw beyond matrix bounds."""
        matrix.bresenham_line(self.set, x0, y0, x1, y1, color, colorFunc)

    # Xiaolin Wu's Line Algorithm
    def wu_line(self, x0, y0, x1, y1, color=None, colorFunc=None):
        matrix.wu_line(self.set, x0, y0, x1, y1, color, colorFunc)

    def drawRect(self, x, y, w, h, color=None, aa=False):
        """Draw rectangle with top-left corner at x,y, width w and height h"""
        matrix.draw_rect(self.set, x, y, w, h, color, aa)

    def fillRect(self, x, y, w, h, color=None, aa=False):
        """Draw solid rectangle with top-left corner at x,y, width w and height h"""
        matrix.fill_rect(self.set, x, y, w, h, color, aa)

    def fillScreen(self, color=None):
        """Fill the matrix with the given RGB color"""
        matrix.fill_rect(self.set, 0, 0, self.width, self.height, color)

    def drawRoundRect(self, x, y, w, h, r, color=None, aa=False):
        """Draw rectangle with top-left corner at x,y, width w, height h, and corner radius r"""
        matrix.draw_round_rect(self.set, x, y, w, h, r, color, aa)

    def fillRoundRect(self, x, y, w, h, r, color=None, aa=False):
        """Draw solid rectangle with top-left corner at x,y, width w, height h, and corner radius r"""
        matrix.fill_round_rect(self.set, x, y, w, h, r, color, aa)

    def drawTriangle(self, x0, y0, x1, y1, x2, y2, color=None, aa=False):
        """Draw triangle with points x0,y0 - x1,y1 - x2,y2"""
        matrix.draw_triangle(self.set, x0, y0, x1, y1, x2, y2, color, aa)

    def fillTriangle(self, x0, y0, x1, y1, x2, y2, color=None, aa=False):
        """Draw solid triangle with points x0,y0 - x1,y1 - x2,y2"""
        matrix.fill_triangle(self.set, x0, y0, x1, y1, x2, y2, color, aa)

    fillTrangle = fillTriangle  # DEPRECATED!

    def drawChar(self, x, y, c, color, bg, aa=False, font=font.default_font, font_scale=1):
        matrix.draw_char(self.fonts, self.set, self.width, self.height,
                         x, y, c, color, bg, aa, font, font_scale)

    def drawText(self, text, x=0, y=0, color=None, bg=colors.Off, aa=False, font=font.default_font, font_scale=1):
        matrix.draw_text(self.fonts, self.set, text, self.width, self.height,
                         x, y, color, bg, aa, font, font_scale)


# This is DEPRECATED
LEDMatrix = Matrix



from . import Rotation, rotate_and_flip
from . matrix import gen_matrix


def gen_cube(dx, dy, dz, xy_serpentine=True, offset=0,
             xy_rotation=Rotation.ROTATE_0, z_rotation=Rotation.ROTATE_0,
             y_flip=False, z_flip=False):
    result = []
    plane_offset = offset
    for z in range(dz):
        plane = gen_matrix(dx, dy, serpentine=xy_serpentine,
                           offset=plane_offset, rotation=xy_rotation,
                           y_flip=y_flip)
        plane_offset += (dx * dy)
        result.append(plane)

    result = rotate_and_flip(result, z_rotation, z_flip)

    return result


def pixel_positions_from_cube(coord_map):
    num = 0
    dx, dy, dz = (0, 0, len(coord_map))

    for xy in coord_map:
        if len(xy) > dx:
            dy = len(xy)
        for x in xy:
            if len(x) > dx:
                dx = len(x)
            num += len(x)

    result = [None] * num
    for z in range(dz):
        for y in range(dy):
            for x in range(dx):
                result[coord_map[z][y][x]] = [x, y, z]
    return result



"""
Run a project description file.
"""

import gitty, json
from . import common_flags, simpixel
from .. import log
from .. util import files


def make_desc(name, is_json):
    if not name:
        return {}

    data = name if is_json else files.opener(name).read()
    try:
        return json.loads(data)
    except ValueError as e:
        e.args += tuple(['in %s' % name])
        raise


def make_animation(name, is_json, args):
    desc = make_desc(name, is_json)
    return common_flags.make_animation(args, desc)


def run(args, settings):
    gitty.sys_path.extend(args.path)
    if args.simpixel:
        simpixel.open_simpixel(args.simpixel)
    elif args.s:
        simpixel.open_simpixel()
    task = make_animation(args.name, args.json, args)
    try:
        task.start()
    except KeyboardInterrupt:
        print('\nTermination requested by user.')
        task.cleanup()


def set_parser(parser):
    parser.set_defaults(run=run)
    parser.description = 'Run specified BiblioPixel project from file or URL.'

    common_flags.add_project_flags(parser)

    parser.add_argument(
        'name', nargs='?',
        help='Path project file - can be a URL or file system location',
        default='')

    parser.add_argument(
        '-j', '--json', action='store_true',
        help='Enter JSON directly as a command line argument.')



from .. project import aliases, project
from . import animation, runner
import traceback
from .. import log


class Collection(animation.BaseAnimation):
    def __init__(self, layout, animations=None, no_fail=False):
        super().__init__(layout)
        self.animations = [self._make_animation(i, no_fail) for i in animations or []]
        self.index = 0
        self.internal_delay = 0  # never wait

    # Override to handle all the animations
    def cleanup(self, clean_layout=True):
        for a in self.animations:
            if a:
                a.cleanup()
        super().cleanup(clean_layout)

    def add_animation(self, anim, **kwds):
        # DEPRECATED.
        anim.set_runner(runner.Runner(**kwds))
        self.animations.append(anim)

    def pre_run(self):
        self.index = -1

    @property
    def current_animation(self):
        if (0 <= self.index < len(self.animations)):
            return self.animations[self.index]
        else:
            return None

    def _make_animation(self, a, no_fail=False):
        if isinstance(a, str):
            desc = {'animation': a}
        elif isinstance(a, dict):
            if 'animation' in a:
                # TODO: this hackiness occurs because this indirectly gets
                # called from different places.  Figure out a better way.
                desc = a
            else:
                desc = {'animation': a}
        else:
            desc = {'animation': a[0], 'run': a[1]}
        desc = aliases.resolve(desc)
        try:
            return project.make_animation(layout=self.layout, **desc)
        except:
            log.error(traceback.format_exc())
            return None



from . circle import Circle
from . matrix import Matrix
from . cube import Cube
from . pov import POV
from . strip import Strip
from . geometry import Rotation
from . geometry.matrix import gen_matrix
from . geometry.circle import gen_circle
from . geometry.cube import gen_cube

# These are DEPRECATED
from . circle import LEDCircle
from . cube import LEDCube
from . matrix import LEDMatrix
from . pov import LEDPOV
from . strip import LEDStrip



import json
from .. project import project

"""Common command line arguments for run and demo."""


def _get_version():
    from os.path import abspath, dirname, join
    filename = join(dirname(dirname(dirname(abspath(
        __file__)))), 'bibliopixel', 'VERSION')
    return open(filename).read().strip()


COMPONENTS = 'driver', 'layout', 'animation'
PRESET_LIBRARY_DEFAULT = '~/.bibliopixel'
LOG_LEVELS = ('debug', 'info', 'warning', 'error', 'critical')
ENABLE_PRESETS = False

PATH_HELP = """\
A list of directories, separated by colons, 'which are added to the end of
`sys.path`.

You can also use gitty-style paths which start with `//git/` to
dynamically load a library from a public git repository.

See https://github.com/ManiacalLabs/BiblioPixel/wiki/BiblioPixel-Paths
for more information.
"""

LOGLEVEL_HELP = """\
Set what level of events to log. Higher log levels print less."""

VERBOSE_HELP = """\
If this is set, then errors are reported with a full stack trace.
If not set, just the exception message is printed.
"""

VERSION = _get_version()
VERSION_HELP = """\
Print the current version number of BiblioPixel (%s).
""" % VERSION

PRESET_HELP = """Filenames for preset library"""


def add_common_flags(parser):
    parser.add_argument(
        '--loglevel', choices=LOG_LEVELS, default='info', help=LOGLEVEL_HELP)
    parser.add_argument(
        '--verbose', '-v', action='store_true', help=VERBOSE_HELP)
    parser.add_argument(
        '--version', action='store_true', help=VERSION_HELP)
    if ENABLE_PRESETS:
        parser.add_argument(
            '--presets', help=PRESET_HELP, default=PRESET_LIBRARY_DEFAULT)


def add_project_flags(parser):
    parser.add_argument(
        '-a', '--animation', default=None,
        help='Default animation type if no animation is specified')

    parser.add_argument(
        '-d', '--driver', default='simpixel',
        help='Default driver type if no driver is specified')

    parser.add_argument(
        '-l', '--layout', default='matrix',
        help='Default layout class if no layout is specified')

    parser.add_argument(
        '-n', '--numpy', action='store_true',
        help='Use numpy if it available.')

    parser.add_argument(
        '-p', '--path', default=None, help=PATH_HELP)

    parser.add_argument(
        '-s', action='store_true', help='Run SimPixel at the default URL')

    parser.add_argument(
        '--simpixel', help='Run SimPixel at a specific URL')

    parser.add_argument(
        '-t', '--ledtype', default=None,
        help='Default LED type if no LED type is specified')


def make_animation(args, desc):
    def get_value(name):
        value = args and getattr(args, name)
        if not value:
            return {}

        if '{' in value:
            return json.loads(value)

        return {'typename': value}

    project_flags = {name: get_value(name) for name in COMPONENTS}
    if args and args.ledtype:
        project_flags['driver']['ledtype'] = args.ledtype

    if args and args.numpy:
        project_flags['maker'] = {'use_numpy': True}

    return project.project_to_animation(desc, project_flags)



import multiprocessing
from .. animation.animation import STATE
from .. animation import collection
from . import server
from .. import log


DEFAULT_OFF = 'OFF_ANIM'
DEFAULT_ANIM_CONFIG = {
    'bgcolor': '#00ff00',
    'font_color': '#ffffff',
    'valid': True
}


OFF_ANIM = {
    'animation': 'bibliopixel.animation.off.OffAnim',
    'run': {'threaded': True}
}


BAD_DEFAULT_ERROR = """\
`{}` is not a valid default!
It must be one of the configured animation names."""


def normalize_name(name):
    return ''.join('_' if e is ' ' else e for e in name if e.isalnum() or e is ' ')


class RemoteControl(collection.Collection):
    def __init__(self, layout, animations,
                 external_access=False, port=5000,
                 title='BiblioPixel Remote', bgcolor='black',
                 font_color='white', default=None):

        self.internal_delay = 0  # never wait

        # normalize name and display first
        for i, anim in enumerate(animations):
            adesc = anim['animation']
            adesc['data'] = dict(DEFAULT_ANIM_CONFIG, **adesc.get('data', {}))
            if not adesc.get('name'):
                log.error('All animations should have the `name` parameter: {}'.format(adesc))
                adesc['name'] = str(i)
            adesc['data']['display'] = adesc['data'].get('display', adesc['name'])
            adesc['name'] = normalize_name(adesc['name'])
            anim['run'] = anim.get('run', {})
            anim['run']['threaded'] = True  # threaded required
        super().__init__(layout, animations, True)

        self.name_map = {}
        self.anim_cfgs = []
        for i, anim in enumerate(self.animations):
            adesc = animations[i]['animation']
            name = adesc['name']  # if failed to load anim will be None
            if anim is None:
                adesc['data']['valid'] = False
                adesc['data']['display'] = 'LOAD FAILED: ' + adesc['data']['display']
                adesc['data']['bgcolor'] = 'rgb(48, 48, 48)'
                adesc['data']['font_color'] = 'white'

            if name in self.name_map:
                raise ValueError('Cannot have multiple animations with the same name: ' + name)
            self.name_map[name] = anim and i
            self.anim_cfgs.append(animations[i]['animation']['data'])
            self.anim_cfgs[-1]['name'] = name
            if anim:
                anim.on_completion = self.on_completion

        if default is None:
            self.default = DEFAULT_OFF
        else:
            self.default = normalize_name(default)
            self.name_map[DEFAULT_OFF] = self.name_map[self.default]
            if self.default is None:
                self.default = DEFAULT_OFF

        if self.default == DEFAULT_OFF:
            self.animations.append(self._make_animation(OFF_ANIM))
            self.name_map[DEFAULT_OFF] = len(self.animations) - 1

        self.index = self.name_map[DEFAULT_OFF]  # start with default animation

        self.ui_config = {
            'bgcolor': bgcolor,
            'font_color': font_color,
            'title': title
        }

        self.q_send = multiprocessing.Queue()
        self.q_recv = multiprocessing.Queue()

        server_args = (
            external_access,
            port,
            self.q_send,
            self.q_recv
        )

        self.server = multiprocessing.Process(target=server.run_server,
                                              args=server_args)

        self.handlers = {
            'run_animation': self.run_animation,
            'stop_animation': self.stop_animation,
            'get_config': self.get_config
        }

    def cleanup(self, clean_layout=True):
        self.q_recv.close()
        self.q_send.close()
        self.server.terminate()
        super().cleanup(clean_layout)

    def on_completion(self, reason):
        if reason != STATE.canceled:
            self.stop_animation()

    # API Handlers
    def run_animation(self, name=None):
        if name is None:
            name = self.default

        if name not in self.name_map:
            return False, 'Invalid animation name: {}'.format(name)

        if self.current_animation:
            self.current_animation.cleanup(clean_layout=False)
            self.index = -1

        log.info('Running animation: {}'.format(name))
        self.index = self.name_map[name]
        self.current_animation.start()
        return True, None

    def stop_animation(self, data):
        return self.run_animation()

    def get_config(self, data):
        resp = {
            'ui': self.ui_config,
            'animations': list(self.anim_cfgs)
        }
        return True, resp

    def pre_run(self):
        self.server.start()

    def step(self, amt=None):
        self.run_animation()
        while not self.threading.stop_event.isSet():
            recv = self.q_recv.get()
            req = recv['req']
            if req not in self.handlers:
                resp = False, '{} is not a valid request!'.format(req)
            else:
                resp = self.handlers[req](recv['data'])
            self.q_send.put(resp)
